from .query_creator import QueryCreator
from .abstract import QueryCreatorABC
from .exeptions import TheInstanceDoesNotExistExeption
from .exeptions import TheMultipleResultsExeption

__all__ = [
    "QueryCreatorABC",
    "QueryCreator",
    "TheInstanceDoesNotExistExeption",
    "TheMultipleResultsExeption"
]



from .abstract import BaseField


class BooleanField(BaseField):
    def __init__(self, field_name=None, null=False, primary=False, unique=False, default=None):
        super().__init__(int, field_name, null, primary, unique, default)

    def get_basic_sql_line(self, sql_type="INTEGER") -> str:
        return f"{self.field_name} {sql_type}"

    def validate(self, value) -> None:
        super().validate(value)
        if value not in (0, 1):
            raise TypeError(
                f"Invalid value for field '{self.field_name}': "
                f"The value {value} cannot be converted to a logical value"
            )



from .abstraction import BaseCommand
from .update_db_command import UpdateDatabaseCommand
from .command_manager import CommandManager


__all__ = [
    'CommandManager',
    'UpdateDatabaseCommand',
    'BaseCommand'
]



from .db.backends import SQLiteBackend, PostgreSQLBackend
import os


_active_backend = None

_registered_backends = {
    "sqlite": SQLiteBackend,
    "postgresql": PostgreSQLBackend
}


def register_backend(name, backend_class):
    if name in _registered_backends:
        raise ValueError(f"Backend '{name}' is already registered.")
    _registered_backends[name] = backend_class


def set_database_backend(backend_name, *args, **kwargs):
    global _active_backend

    if (backend := _registered_backends.get(backend_name)):
        _active_backend = backend(*args, **kwargs)
        _active_backend.connect(*args, **kwargs)
    else:
        raise ValueError(f"Unsupported backend: {backend_name}")


def _get_active_backend(config_file="dbeasyorm.ini"):
    if _active_backend is None:
        try:
            import configparser

            if config_file and os.path.exists(config_file):
                config_parser = configparser.ConfigParser()
                config_parser.read(config_file)
                if (db_config := dict(config_parser["database"])):
                    db_type = db_config.get("db_type")
                    if db_type is None:
                        raise ValueError("Database type ('db_type') is missing in the configuration.")

                    set_database_backend(db_type, **db_config)
                    return _active_backend
                raise RuntimeError(f"No database backend has been configured in {config_file}.")
        except ImportError:
            pass

        raise RuntimeError("No database backend has been configured.")
    return _active_backend


def _get_folders_for_migration_search(config_file="dbeasyorm.ini"):
    try:
        import configparser

        if config_file and os.path.exists(config_file):
            config_parser = configparser.ConfigParser()
            config_parser.read(config_file)
            if (app_config := dict(config_parser["app"])):
                app_dir = app_config.get("dir")
                if app_dir is None:
                    raise ValueError(
                        "The specified folder for your application in which we will search for migrations is missing."
                        "Please set (dir) key = value"
                    )

                return app_dir
            raise RuntimeError(f"No application has been configured in {config_file}.")
    except ImportError:
        raise RuntimeError("Please install `configparser`")



import argparse
import os

from dbeasyorm.migrations import MigrationHandler
from .abstraction import BaseCommand
from dbeasyorm.config import _get_folders_for_migration_search


class UpdateDatabaseCommand(BaseCommand):
    """Concrete implementation of the 'update-database' command."""

    def name(self) -> str:
        return "update-database"

    def help(self) -> str:
        return "Update the database."

    def configure_arguments(self, parser: argparse.ArgumentParser) -> None:
        parser.add_argument("-l", "--loockup-folder", type=str, help="Path to the lookup folder")
        parser.add_argument("-i", "--id-migrations", type=str, help="ID of specific migrations")
        parser.add_argument("-r", "--restore", action="store_true", help="Restore database to the previous state")
        parser.add_argument("-c", "--config", type=str, help="Path to the config.ini file")

    def handle(self, loockup_folder=None, id_migrations=None, restore=False, config_file=None, **kwargs) -> None:
        loockup_folder = os.path.join(os.getcwd(), _get_folders_for_migration_search()) if not loockup_folder else loockup_folder
        mig_handler = MigrationHandler(config_file)
        return mig_handler.update_database(
            loockup_folder=loockup_folder,
            id_migrations=id_migrations,
            restore=restore
        )



from .abstract import BaseField


class FloatField(BaseField):
    def __init__(self, field_name=None, null=False, primary=False, unique=False, default=None):
        super().__init__(float, field_name, null, primary, unique, default=None)

    def get_basic_sql_line(self, sql_type="REAL") -> str:
        return f'{self.field_name} {sql_type}'



from .abstract import BaseField


class ForeignKey(BaseField):
    def __init__(self, related_model, field_name=None, null=True, primary=False, unique=False, on_delete="CASCADE"):
        super().__init__(int, field_name, null, primary, unique)
        self.related_model = related_model
        self.on_delete = on_delete

    def get_sql_line(self, db_backend_constrain_gen=None) -> str:
        return db_backend_constrain_gen(self.field_name, self.related_model.__name__.upper(), self.on_delete)

    def get_basic_sql_line(self, sql_type="INTEGER") -> str:
        return f'{self.field_name} {sql_type}'

    def validate(self, value):
        super().validate(value)
        if not isinstance(value, int):
            raise TypeError(
                f"Invalid value for ForeignKey field '{self.field_name}': expected int, got {type(value).__name__}."
            )



from abc import ABC, abstractmethod
from dbeasyorm.fields import BaseField


class DataBaseBackend(ABC):

    @abstractmethod
    def get_placeholder(self) -> str:
        ...

    @abstractmethod
    def get_sql_type(self, type: type) -> str:
        ...

    @abstractmethod
    def get_sql_types_map(self) -> dict:
        ...

    @abstractmethod
    def get_foreign_key_constraint(self, field_name, related_table, on_delete) -> str:
        ...

    @abstractmethod
    def connect(self, *args, **kwargs) -> object:
        ...

    @abstractmethod
    def execute(self, query: str, params=None):
        ...

    def generate_insert_sql(self, table_name: str, columns: tuple) -> str:
        ...

    @abstractmethod
    def generate_select_sql(self, table_name: str, columns: tuple, where_clause: tuple, limit: int = None, offset: int = None) -> str:
        ...

    @abstractmethod
    def generate_join_sql(self, table_name: str, on: str, join_type: str) -> str:
        ...

    @abstractmethod
    def generate_update_sql(self, table_name: str, set_clause: tuple, where_clause: tuple) -> str:
        ...

    @abstractmethod
    def generate_delete_sql(self, table_name: str, where_clause: tuple) -> str:
        ...

    @abstractmethod
    def generate_create_table_sql(self, table_name: str, fields: BaseField) -> str:
        ...

    @abstractmethod
    def generate_alter_field_sql(self, *args, **kwargs) -> str:
        ...

    @abstractmethod
    def generate_drop_field_sql(self, *args, **kwargs) -> str:
        ...

    @abstractmethod
    def generate_drop_table_sql(self, table_name: str) -> str:
        ...

    @abstractmethod
    def get_database_schemas(self) -> dict:
        ...

    def get_sql_val_repr(self, value):
        return f"'{value}'" if isinstance(value, str) else ("NULL" if value is None else f"{value}")



class BaseFieldMeta(type):
    def __new__(cls, name, bases, attrs):
        constraints = {}
        for attr_name, attr_value in attrs.items():
            if attr_name != "field_name":
                constraints[attr_name] = attr_value

        attrs['_constraints'] = constraints
        return super().__new__(cls, name, bases, attrs)


class BaseField(metaclass=BaseFieldMeta):
    def __init__(self, python_type, field_name=None, null=False, primary=False, unique=False, autoincrement=False, default=None):
        self.python_type = python_type
        self.field_name = field_name
        self.null = null
        self.primary = primary
        self.unique = unique
        self.autoincrement = autoincrement
        self.default = default
        self.value = self.default if self.default else None

    def __repr__(self):
        return f"<Field field_name={self.field_name} {self.__class__.__name__}>"

    def get_sql_line(self, *args, **kwargs) -> str:
        sql_line = self.get_basic_sql_line(**kwargs)
        if self.primary:
            sql_line += ' PRIMARY KEY'
        if self.autoincrement and self.primary:
            sql_line += ' AUTOINCREMENT'
        if self.null is False and not self.primary:
            sql_line += ' NOT NULL'
        if self.unique and not self.primary:
            sql_line += ' UNIQUE'

        return sql_line

    def get_basic_sql_line(self, sql_type: str) -> str:
        raise NotImplementedError("This method must be implemented in subclasses")

    def validate(self, value):
        if value is not None and not isinstance(value, self.python_type):
            raise TypeError(
                f"Invalid value for field '{self.field_name}': "
                f"expected {self.python_type.__name__}, got {type(value).__name__}."
            )

        if value is None and not self.null:
            raise ValueError(
                f"Invalid value for field '{self.field_name}': "
                f"the value is expected, but got None."
            )



import sqlite3
from .abstract import DataBaseBackend
from dbeasyorm.fields import BaseField, ForeignKey


class SQLiteBackend(DataBaseBackend):
    def __init__(self, database_path: str, *args, **kwargs):
        self.database_path = database_path
        self.cursor = None
        self.connection = None
        self.type_map = self.get_sql_types_map()

    def get_placeholder(self) -> str:
        return "?"

    def get_sql_type(self, type) -> str:
        return self.type_map.get(type)

    def get_sql_types_map(self) -> dict:
        return {
            int: "INTEGER",
            float: "REAL",
            bytes: "BLOB",
            bool: "INTEGER",
            str: "TEXT"
        }

    def get_foreign_key_constraint(self, field_name: str, related_table: str, on_delete: str) -> str:
        return (
            f"{field_name} INTEGER ",
            f"FOREIGN KEY ({field_name}) REFERENCES {related_table} (_id) "
            f"ON DELETE {on_delete}"
        )

    def connect(self, **kwargs) -> DataBaseBackend:
        self.connection = sqlite3.connect(self.database_path)
        self.cursor = self.connection.cursor()
        return self

    def execute(self, query: str, params=None) -> sqlite3.Cursor:
        # Split the query into individual statements and execute them
        statements = query.strip().split(";")
        for statement in statements:
            if statement.strip():
                self.cursor.execute(statement.strip(), params or ())
        self.connection.commit()
        return self.cursor

    def generate_insert_sql(self, table_name: str, columns: tuple) -> str:
        columns_str = ', '.join(columns)
        placeholders = ', '.join([self.get_placeholder() for _ in columns])
        return f"INSERT INTO {table_name} ({columns_str}) VALUES ({placeholders})"

    def generate_select_sql(self, table_name: str, columns: tuple, where_clause: dict = None, limit: int = None, offset: int = None) -> str:
        where_sql = ""
        if where_clause:
            where_sql = " WHERE " + " AND ".join([f"{col} = " + self.get_sql_val_repr(val) for col, val in where_clause.items()])

        limit_offset_sql = ""
        if limit is not None:
            limit_offset_sql = f" LIMIT {limit}"
        if offset is not None:
            limit_offset_sql += f" OFFSET {offset}"

        return f"SELECT {', '.join(columns) if columns else f'{table_name}.*'} FROM {table_name}{where_sql}{limit_offset_sql}"

    def generate_join_sql(self, table_name: str, on: str, join_type: str) -> str:
        return f" {join_type} JOIN {table_name} ON {on}"

    def generate_update_sql(self, table_name: str, set_clause: tuple, where_clause: tuple):
        set_sql = ', '.join([f"{col}={self.get_placeholder()}" for col in set_clause])
        where_sql = " AND ".join([f"{col}={self.get_placeholder()}" for col in where_clause]) if where_clause else ""
        return f"UPDATE {table_name} SET {set_sql} WHERE {where_sql}"

    def generate_delete_sql(self, table_name: str, where_clause: tuple):
        where_sql = " AND ".join([f"{col}={self.get_placeholder()}" for col in where_clause]) if where_clause else ""
        return f"DELETE FROM {table_name} WHERE {where_sql}"

    def generate_create_table_sql(self, table_name: str, fields: BaseField):
        columns = []
        foreign_keys = []

        for field in fields:
            if isinstance(field, ForeignKey):
                column, foreign_key = field.get_sql_line(self.get_foreign_key_constraint)
                columns.append(column)
                foreign_keys.append(foreign_key)
            else:
                columns.append(
                    field.get_sql_line(sql_type=self.get_sql_type(field.python_type))
                )
        table_body = ", \n".join(columns + foreign_keys)
        return f"""CREATE TABLE IF NOT EXISTS {table_name} ({table_body});"""

    def generate_alter_field_sql(self, model: BaseField, db_columns: dict, *args, **kwargs) -> str:
        table_name = model.query_creator.get_table_name()
        sql_result = ''
        db_columns = ", ".join(db_columns.keys())

        # sql_create_new_table_query
        sql_result += self.generate_create_table_sql(f"{table_name}_NEW", list(model._fields.values()))
        sql_result += f"""INSERT INTO {table_name}_NEW ({db_columns}) SELECT {db_columns} FROM {table_name};"""
        sql_result += self.generate_drop_table_sql(table_name=table_name)
        sql_result += f"ALTER TABLE {table_name}_NEW RENAME TO {table_name};"
        return sql_result

    def generate_drop_field_sql(self, model: BaseField, *args, **kwargs) -> str:
        table_name = model.query_creator.get_table_name()
        sql_result = ''
        columns = ", ".join(model._fields.keys())

        # sql_create_new_table_query
        sql_result += self.generate_create_table_sql(f"{table_name}_NEW", list(model._fields.values()))
        sql_result += f"""INSERT INTO {table_name}_NEW ({columns}) SELECT {columns} FROM {table_name};"""
        sql_result += self.generate_drop_table_sql(table_name=table_name)
        sql_result += f"ALTER TABLE {table_name}_NEW RENAME TO {table_name};"
        return sql_result

    def generate_drop_table_sql(self, table_name: str) -> str:
        return f"DROP TABLE {table_name};"

    def get_database_schemas(self) -> dict:
        schema = {}

        self.cursor.execute("SELECT name FROM sqlite_master WHERE type='table';")
        tables = self.cursor.fetchall()
        for table in tables:
            table_name = table[0]
            if table_name == 'sqlite_sequence':
                continue

            self.cursor.execute(f"PRAGMA table_info({table_name});")
            columns = self.cursor.fetchall()
            schema[table_name] = {col[1]: col[2] for col in columns}

        return schema



import time


class QueryCounter:
    __queries = []
    __start_registration = False

    @classmethod
    def start_registration(cls):
        cls.__start_registration = True

    @classmethod
    def end_registration(cls):
        cls.__start_registration = False

    @classmethod
    def register_query(cls, sql):
        if cls.__start_registration:
            cls.__queries.append(sql)

    @classmethod
    def clear_queries(cls):
        cls.__queries = []

    @classmethod
    def get_queries(cls):
        return cls.__queries

    @classmethod
    def get_query_count(cls):
        return len(cls.__queries)

    def __enter__(self):
        self.clear_queries()
        self.start_time = time.time()
        self.start_registration()
        print("Query logging started. Previous logs cleared.")
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.end_registration()
        elapsed_time = time.time() - self.start_time
        print(f"Query logging ended. Total queries: {self.get_query_count()}")
        print("Queries: ")
        [print(query) for query in self.get_queries()]
        print(f"Elapsed time: {elapsed_time:.2f} seconds.")
        if exc_type:
            print(f"Error occurred: {exc_value}")



from abc import ABC, abstractmethod


class QueryCreatorABC(ABC):

    @abstractmethod
    def execute(self) -> object:
        ...

    @abstractmethod
    def create(self, *args, **kwargs) -> object:
        ...

    @abstractmethod
    def update(self, *args, **kwargs) -> object:
        ...

    @abstractmethod
    def delete(self, *args) -> object:
        ...

    @abstractmethod
    def all(self) -> object:
        ...

    @abstractmethod
    def get_one(self, *args, **kwargs) -> object:
        ...

    @abstractmethod
    def filter(self, *args, **kwargs) -> object:
        ...



class TheInstanceDoesNotExistExeption(Exception):
    def __init__(self, message="The instance does not exist", code=None, instance_class=None):
        self.message = message
        self.code = code
        self.instance_class = instance_class
        super().__init__(self.message)

    def __str__(self):
        if self.code:
            return f"[Error {self.code}]: {self.message}"
        if self.instance_class:
            return f"The instance of {self.instance_class} does not exist"
        return self.message


class TheMultipleResultsExeption(Exception):
    def __init__(self, message="Expected one result, received many", code=None, instance_class=None):
        self.message = message
        self.code = code
        self.instance_class = instance_class
        super().__init__(self.message)

    def __str__(self):
        if self.code:
            return f"[Error {self.code}]: {self.message}"
        if self.instance_class:
            return f"Expected one result of {self.instance_class}, received many"
        return self.message



from colorama import Fore
from dbeasyorm.db.backends import DataBaseBackend, SQLiteBackend

from .messages import print_success, print_info, print_line


class MigrationExecutor:
    def __init__(self, db_backend: DataBaseBackend):
        self.db_backend = db_backend
        self.sql = ""

    def execute_detected_migration(self, detected_migration: dict) -> None:
        info_str = "Detected migrations: " + str(detected_migration)
        print_line(Fore.BLUE, '-')
        print_info(info_str)
        print_line(Fore.BLUE, '-')

        self.sql = self.append_table_creation_sql(detected_migration.get('create_tables'), self.sql)
        self.sql = self.append_delete_tables_sql(detected_migration.get('drop_tables'), self.sql)
        self.sql = self.append_insert_cols_sql(detected_migration.get('add_columns'), self.sql)
        self.sql = self.append_delete_cols_sql(detected_migration.get('remove_columns'), self.sql)

        self.db_backend.execute(query=self.sql)
        print_success("All database migrations applied")

    def append_table_creation_sql(self, models: list, sql: str) -> str:
        for model in models:
            model.migrate()
            sql += model.query_creator.sql
            sql += " \n"
        return sql

    def append_delete_tables_sql(self, columns: list, sql: str) -> str:
        for table_name in columns:
            sql += self.db_backend.generate_drop_table_sql(table_name=table_name)
        return sql

    def append_insert_cols_sql(self, columns: list, sql: str) -> str:
        if isinstance(self.db_backend, SQLiteBackend):
            unique_dict = {item[0]: (item[0], item[1], item[2], item[3]) for item in columns}
            unique_list = list(unique_dict.values())
            for _, _, model, db_columns in unique_list:
                sql += self.db_backend.generate_alter_field_sql(model=model, db_columns=db_columns)
        else:
            for table_name, col, _, _ in columns:
                sql += self.db_backend.generate_alter_field_sql(table_name=table_name, field=col)
        return sql

    def append_delete_cols_sql(self, columns: list, sql: str) -> str:
        if isinstance(self.db_backend, SQLiteBackend):
            unique_dict = {item[0]: (item[0], item[1], item[2]) for item in columns}
            unique_list = list(unique_dict.values())
            for _, _, model in unique_list:
                sql += self.db_backend.generate_drop_field_sql(model=model)
        else:
            for table_name, col, _, _ in columns:
                sql += self.db_backend.generate_drop_field_sql(table_name=table_name, field=col)
        return sql



from .abstract import QueryCreatorABC
from dbeasyorm.config import _get_active_backend
from dbeasyorm.fields import ForeignKey
from dbeasyorm.models.exeptions import TheKeyIsNotAForeignKeyError
from .exeptions import TheInstanceDoesNotExistExeption
from .query_counter import QueryCounter
from .utils import insert_pattern_into_str


class QueryCreator(QueryCreatorABC):
    query_counter = QueryCounter()

    def __init__(self, model):
        super().__init__()
        self.backend = _get_active_backend()
        self.model = model
        self.cursor = None
        self.cnx = None
        self.setup_new_query_params()

    def setup_new_query_params(self):
        self.sql = ''
        self.many = False
        self.values = None
        self.return_id = False

    def execute(self) -> object:
        self.query_counter.register_query(self.sql)

        cursor = self.backend.execute(query=self.sql, params=self.values)
        return cursor.lastrowid if self.return_id else \
            (self._fetch_all(cursor) if self.many else self._fetch_one(cursor))

    def _fetch_one(self, cursor) -> object:
        data = cursor.fetchone()
        if not data:
            raise TheInstanceDoesNotExistExeption(instance_class=self.get_class_name())
        return self._map_row_to_object(data)

    def _fetch_all(self, cursor) -> list:
        rows = cursor.fetchall()
        return [self._map_row_to_object(row) for row in rows]

    def _map_row_to_object(self, row) -> object:
        column_names = [desc[0] for desc in self.backend.cursor.description]

        if not hasattr(self, "join_tables"):
            data = dict(zip(column_names, row))
            return self.model(**data)

        return self._map_row_to_object_with_relations(column_names, row)

    def _map_row_to_object_with_relations(self, column_names: list, row: list) -> object:
        base_model_data = {}
        related_model_data = {field: {} for field in self.join_tables.keys()}
        table_prefix_map = {data["table"]: field_name for field_name, data in self.join_tables.items()}

        for column, value in zip(column_names, row):
            table_prefix = next((prefix for prefix in table_prefix_map if column.startswith(prefix)), None)

            if table_prefix:
                related_field_name = table_prefix_map[table_prefix]
                field_name = column[len(table_prefix) + 1:]
                related_model_data[related_field_name][field_name] = value
            else:
                base_model_data[column] = value

        for related_field_name, data in self.join_tables.items():
            related_model_data[related_field_name] = data["model"](**related_model_data[related_field_name])

        base_model_data.update(related_model_data)
        delattr(self, 'join_tables')
        return self.model(**base_model_data)

    def get_table_name(self) -> str:
        return self.get_class_name().__name__.upper()

    def get_class_name(self) -> str:
        if isinstance(self.model, type):
            return self.model
        return self.model.__class__

    def create(self, *args, **kwargs) -> QueryCreatorABC:
        self.setup_new_query_params()
        self.sql = self.backend.generate_insert_sql(
            table_name=self.get_table_name(),
            columns=kwargs.keys()
        )
        self.return_id = True
        self.values = tuple(kwargs.values())
        return self

    def update(self, where: dict = None, *args, **kwargs) -> QueryCreatorABC:
        self.setup_new_query_params()
        self.sql = self.backend.generate_update_sql(
            table_name=self.get_table_name(),
            set_clause=tuple(kwargs.keys()),
            where_clause=tuple(where.keys())
        )
        self.return_id = True
        self.values = tuple(list(kwargs.values()) + list(where.values()))
        return self

    def delete(self, where: dict = None) -> QueryCreatorABC:
        self.setup_new_query_params()
        self.sql = self.backend.generate_delete_sql(
            table_name=self.get_table_name(),
            where_clause=tuple(where.keys())
        )
        self.values = tuple(where.values())
        self.return_id = True
        return self

    def all(self, *args, **kwargs) -> QueryCreatorABC:
        return self.filter(*args)

    def filter(self, *args, **kwargs) -> QueryCreatorABC:
        self.setup_new_query_params()
        self.sql = self.backend.generate_select_sql(
            table_name=self.get_table_name(),
            columns=kwargs.get('columns', None),
            where_clause=kwargs
        )
        self.many = True
        return self

    def join(self, field_name: str, on: str = None, join_type: str = "INNER"):
        related_field_name = f"id_{field_name}"
        field_instance = self.model._fields.get(related_field_name)
        if not isinstance(field_instance, ForeignKey):
            raise TheKeyIsNotAForeignKeyError()

        join_table_name = field_instance.related_model.query_creator.get_table_name()

        if not hasattr(self, "join_tables"):
            self.join_tables = dict()
        self.join_tables[field_name] = {"table": join_table_name, "model": field_instance.related_model}

        alias_join_table_columns = ', '.join([f"{join_table_name}.{col} AS {join_table_name}_{col}" for col in field_instance.related_model._fields.keys()])

        self.sql = insert_pattern_into_str(r"(?i)(?=\bFROM\b)", f", {alias_join_table_columns} ", self.sql)
        self.sql += self.backend.generate_join_sql(
            join_table_name,
            on=f"{self.get_table_name()}.{related_field_name} = {join_table_name}._id" if not on else on,
            join_type=join_type
        )
        return self

    def get_one(self, *args, **kwargs) -> QueryCreatorABC:
        self.filter(*args, **kwargs)
        self.many = False
        return self

    def migrate_table(self, *args, **kwargs) -> QueryCreatorABC:
        self.sql = self.backend.generate_create_table_sql(
            table_name=self.get_table_name(),
            fields=list(kwargs.values())
        )
        return self



from .migration import Migration
from .migration_detecter import MigrationDetecter
from .migration_handler import MigrationHandler
from .migration_executor import MigrationExecutor

__all__ = [
    'Migration',
    'MigrationDetecter',
    'MigrationHandler',
    'MigrationExecutor',
]



from .abstract import BaseField


class ByteField(BaseField):
    def __init__(self, field_name=None, null=False, primary=False, unique=False, default=None):
        super().__init__(bytes, field_name, null, primary, unique, default)

    def get_basic_sql_line(self, sql_type="BLOB") -> str:
        return f"{self.field_name} {sql_type}"

    def validate(self, value) -> None:
        super().validate(value)
        try:
            value.decode()
        except (UnicodeDecodeError, AttributeError):
            raise TypeError(
                 f"Invalid value for field '{self.field_name}': "
                 f"The {value} is not the byte-like object")



from .abstract import BaseField
from .boolean_field import BooleanField
from .byte_field import ByteField
from .float_field import FloatField
from .integer_filed import IntegerField
from .text_field import TextField
from .foreign_key import ForeignKey


__all__ = [
    "BaseField",
    "BooleanField",
    "ByteField",
    "FloatField",
    "IntegerField",
    "TextField",
    "ForeignKey"
]



from dbeasyorm.fields import IntegerField, BaseField, ForeignKey
from dbeasyorm.query.query_creator import QueryCreator

from .abstract import ModelABC
from .exeptions import ThePrimaryKeyIsImmutable


class ModelMeta(type):
    def __new__(cls, name, bases, attrs):
        fields = {'_id': IntegerField(field_name='_id', null=False, primary=True, autoincrement=True)}

        for attr_name, attr_value in list(attrs.items()):
            if isinstance(attr_value, BaseField):
                if isinstance(attr_value, ForeignKey):
                    attrs, fields = cls._set_property_for_foreign_key_field(attr_name, attr_value, attrs, fields, cls)
                else:
                    fields[attr_name] = attr_value
                    attr_value.field_name = attr_name

        attrs['_fields'] = fields
        new_cls = super().__new__(cls, name, bases, attrs)
        new_cls.query_creator = QueryCreator(new_cls)
        return new_cls

    @staticmethod
    def _set_property_for_foreign_key_field(attr_name: str, attr_value: ForeignKey, attrs: dict, fields: dict, cls: object) -> tuple:
        attrs, fields = cls._append_addional_fields_for_manage_relations(attr_name, attr_value, attrs, fields, cls)

        def getter(self, attr_name=attr_name, related_model=attr_value.related_model):
            id_field_name, model_object_field = cls._generate_additional_fields_for_relations(attr_name)
            if getattr(self, model_object_field, None) is None:
                id_value = getattr(self, id_field_name, None)
                if id_value is not None:
                    setattr(self, model_object_field, related_model.query_creator.get_one(_id=id_value).execute())
            return getattr(self, model_object_field, None)

        def setter(self, value, attr_name=attr_name, related_model=attr_value.related_model):
            if value is None:
                return None

            if not isinstance(value, related_model):
                raise TypeError(f"Expected instance of {related_model.__name__}, got {type(value).__name__}")
            id_field_name, model_object_field = cls._generate_additional_fields_for_relations(attr_name)
            setattr(self, id_field_name, value.id)
            setattr(self, model_object_field, value)

        attrs[attr_name] = property(getter, setter)
        return attrs, fields

    @staticmethod
    def _append_addional_fields_for_manage_relations(attr_name: str, attr_value: ForeignKey, attrs: dict, fields: dict, cls: object) -> tuple:
        id_field_name, model_object_field = cls._generate_additional_fields_for_relations(attr_name)

        attr_value.field_name = id_field_name
        fields[id_field_name] = attr_value
        attrs[id_field_name] = None
        attrs[model_object_field] = None

        return attrs, fields

    @staticmethod
    def _generate_additional_fields_for_relations(attr_name: str) -> tuple:
        id_field_name = f"id_{attr_name}"
        model_object_field = f"{attr_name}_object"
        return id_field_name, model_object_field


class Model(ModelABC, metaclass=ModelMeta):
    def __init__(self, **kwargs):
        for field_name, field_instance in self._fields.items():
            if field_instance.default is not None and kwargs.get(field_name) is None:
                setattr(self, field_name, field_instance.default)
            else:
                setattr(self, field_name, kwargs.get(field_name))
        for field_name, value in kwargs.items():
            if field_name not in list(self._fields.keys()):
                setattr(self, field_name, value)
        self._id = kwargs.get('_id', -1)

    @property
    def id(self) -> int:
        return self._id

    @id.setter
    def id(self, value) -> None:
        if self._id != -1:
            raise ThePrimaryKeyIsImmutable()
        self._id = value

    @property
    def _dict_fields(self) -> dict:
        return {field_name: getattr(self, field_name, None) for field_name in self._fields.keys() if field_name != "_id"}

    def save(self) -> QueryCreator:
        for field_name, field_instance in self._fields.items():
            value = getattr(self, field_name, None)
            field_instance.validate(value)

        if self.id == -1:
            return self.query_creator.create(**self._dict_fields)
        return self.query_creator.update(where=dict(_id=self.id), **self._dict_fields)

    def delete(self) -> QueryCreator:
        return self.query_creator.delete(where=dict(_id=self.id))

    @classmethod
    def migrate(cls) -> QueryCreator:
        return cls.query_creator.migrate_table(**cls._fields)



import ast
import os
import importlib.util
from typing import List, Tuple


class SubclassFinder(ast.NodeVisitor):
    def __init__(self, base_class_name):
        self.base_class_name = base_class_name
        self.subclasses = []

    def visit_ClassDef(self, node):
        if self._is_subclass_of_model(node):
            self.subclasses.append(node.name)
        self.generic_visit(node)

    def _is_subclass_of_model(self, node):
        for base in node.bases:
            if isinstance(base, ast.Name) and base.id == self.base_class_name:
                return True
        return False


class ModelClassesLoader:
    MODEL_BASE_CLASS_NAME = 'Model'

    def load_models(self, loockup_folder: str) -> List[object]:
        classInstances = list()
        detected_subclasses = self._get_detected_subclasses(loockup_folder, self.MODEL_BASE_CLASS_NAME)

        for file_path, class_name in detected_subclasses:
            class_module = self._import_file_as_module(file_path, class_name)
            classInstance = getattr(class_module, class_name, None)()
            classInstances.append(classInstance)
        return classInstances

    def _get_detected_subclasses(self, folder_path: str, base_class_name: str) -> List[Tuple[str, str]]:
        subclasses = []
        for root, _, files in os.walk(folder_path):
            for file in files:
                if file.endswith(".py"):
                    file_path = os.path.join(root, file)
                    subclasses_in_file = self._find_model_subclasses_in_file(file_path, base_class_name)
                    subclasses.extend((file_path, cls) for cls in subclasses_in_file)
        return subclasses

    def _find_model_subclasses_in_file(self, file_path: str, base_class_name: str) -> List[str]:
        with open(file_path, "r", encoding="utf-8") as file:
            tree = ast.parse(file.read(), filename=file_path)
        finder = SubclassFinder(base_class_name)
        finder.visit(tree)
        return finder.subclasses

    def _import_file_as_module(self, file_path: str, module_name: str) -> object:
        spec = importlib.util.spec_from_file_location(module_name, file_path)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        return module



from .abstract import BaseField


class IntegerField(BaseField):
    def __init__(self, field_name=None, null=False, primary=False, autoincrement=False, unique=False, min=None, max=None, default=None):
        super().__init__(int, field_name, null, primary, unique, autoincrement, default=None)
        self.min = min
        self.max = max

    def get_basic_sql_line(self, sql_type="INTEGER") -> str:
        return f'{self.field_name} {sql_type}'

    def validate(self, value) -> None:
        super().validate(value)
        if self.min is not None and value < self.min:
            raise TypeError(f"Value for field '{self.field_name}' is less than {self.min}.")
        if self.max is not None and value > self.max:
            raise TypeError(f"Value for field '{self.field_name}' exceeds {self.max}.")



from colorama import Fore, Style, init
import shutil


init(autoreset=True)


def print_line(color, char="=", style=Style.NORMAL):
    terminal_width = shutil.get_terminal_size().columns
    line = char * terminal_width
    print(f"{color}{style}{line}{Style.RESET_ALL}")


def print_error(message):
    print(f"{Fore.RED}[ERROR] {message}{Style.RESET_ALL}")


def print_success(message):
    print(f"{Fore.GREEN}[SUCCESS] {message}{Style.RESET_ALL}")


def print_info(message):
    print(f"{Fore.CYAN}[INFO] {message}{Style.RESET_ALL}")


def print_warning(message):
    print(f"{Fore.YELLOW}[WARNING] {message}{Style.RESET_ALL}")



from dbeasyorm.commands import CommandManager, UpdateDatabaseCommand


def main():
    manager = CommandManager()

    # Register commands
    manager.register_command(UpdateDatabaseCommand)

    manager.run()


if __name__ == "__main__":
    main()



import psycopg2
from .abstract import DataBaseBackend
from dbeasyorm.fields import BaseField, ForeignKey


class PostgreSQLBackend(DataBaseBackend):
    def __init__(self, host, database: str, user: str, password: str, port=5432, *args, **kwargs):
        self.host = host
        self.database = database
        self.user = user
        self.password = password
        self.port = port
        self.cursor = None
        self.connection = None
        self.type_map = self.get_sql_types_map()

    def get_placeholder(self) -> str:
        return "%s"

    def get_sql_type(self, type) -> str:
        return self.type_map.get(type)

    def get_sql_types_map(self) -> dict:
        return {
            int: "INTEGER",
            float: "DOUBLE PRECISION",
            bytes: "BYTEA",
            bool: "BOOLEAN",
            str: "VARCHAR"
        }

    def connect(self, **kwargs) -> DataBaseBackend:
        self.connection = psycopg2.connect(
            host=self.host, database=self.database, user=self.user, password=self.password, port=self.port
        )
        self.cursor = self.connection.cursor()
        return self

    def get_foreign_key_constraint(self, field_name: str, related_table: str, on_delete: str) -> str:
        return (
            f"{field_name} INTEGER, "
            f"CONSTRAINT fk_{field_name}_to_{related_table} "
            f"FOREIGN KEY ({field_name}) REFERENCES {related_table} (_id) "
            f"ON DELETE {on_delete}"
        )

    def execute(self, query: str, params=None) -> psycopg2.extensions.cursor:
        self.cursor.execute(query, params or ())
        self.connection.commit()
        return self.cursor

    def generate_insert_sql(self, table_name: str, columns: tuple) -> str:
        columns_str = ', '.join(columns)
        placeholders = ', '.join([self.get_placeholder() for _ in columns])
        return f"INSERT INTO {table_name} ({columns_str}) VALUES ({placeholders}) RETURNING id"

    def generate_select_sql(self, table_name: str, columns: tuple, where_clause: dict = None, limit: int = None, offset: int = None) -> str:
        where_sql = ""
        if where_clause:
            where_sql = " WHERE " + " AND ".join([f"{col} = " + self.get_sql_val_repr(val) for col, val in where_clause.items()])

        limit_offset_sql = ""
        if limit is not None:
            limit_offset_sql = f" LIMIT {limit}"
        if offset is not None:
            limit_offset_sql += f" OFFSET {offset}"

        return f"SELECT {', '.join(columns) if columns else '*'} FROM {table_name}{where_sql}{limit_offset_sql}"

    def generate_join_sql(self, table_name: str, on: str, join_type: str) -> str:
        return f" {join_type} JOIN {table_name} ON {on}"

    def generate_update_sql(self, table_name: str, set_clause: tuple, where_clause: tuple):
        set_sql = ', '.join([f"{col} = {self.get_placeholder()}" for col in set_clause])
        where_sql = " AND ".join([f"{col} = {self.get_placeholder()}" for col in where_clause]) if where_clause else ""
        return f"UPDATE {table_name} SET {set_sql} WHERE {where_sql} RETURNING *"

    def generate_delete_sql(self, table_name: str, where_clause: tuple):
        where_sql = " AND ".join([f"{col} = {self.get_placeholder()}" for col in where_clause]) if where_clause else ""
        return f"DELETE FROM {table_name} WHERE {where_sql} RETURNING *"

    def generate_create_table_sql(self, table_name: str, fields: BaseField):
        columns = []
        foreign_keys = []

        for field in fields:
            if isinstance(field, ForeignKey):
                foreign_keys.append(
                    field.get_sql_line(self.get_foreign_key_constraint)
                )
            else:
                columns.append(
                    field.get_sql_line(sql_type=self.get_sql_type(field.python_type))
                )
        table_body = ", \n".join(columns + foreign_keys)
        return f"""CREATE TABLE IF NOT EXISTS {table_name} ({table_body});"""

    def generate_alter_field_sql(self, table_name: str, field: BaseField, *args, **kwargs) -> str:
        if isinstance(field, ForeignKey):
            field_sql = field.get_sql_line(self.get_foreign_key_constraint)
        else:
            field_sql = field.get_sql_line(sql_type=self.get_sql_type(field.python_type))
        return f"ALTER TABLE {table_name} ADD {field_sql};"

    def generate_drop_field_sql(self, table_name: str, field: BaseField, *args, **kwargs) -> str:
        return f"ALTER TABLE {table_name} DROP COLUMN {field};"

    def generate_drop_table_sql(self, table_name: str) -> str:
        return f"DROP TABLE {table_name}"

    def get_database_schemas(self) -> dict:
        schema = {}

        self.cursor.execute("SELECT name FROM sqlite_master WHERE type='table';")
        tables = self.cursor.fetchall()
        for table in tables:
            table_name = table[0]
            if table_name == 'sqlite_sequence':
                continue

            self.cursor.execute(f"PRAGMA table_info({table_name});")
            columns = self.cursor.fetchall()
            schema[table_name] = {col[1]: col[2] for col in columns}

        return schema



class ThePrimaryKeyIsImmutable(Exception):
    def __init__(self, message="The primary key is immutable", code=None):
        self.message = message
        self.code = code
        super().__init__(self.message)

    def __str__(self):
        if self.code:
            return f"[Error {self.code}]: {self.message}"
        return self.message


class TheKeyIsNotAForeignKeyError(Exception):
    def __init__(self, message="The key is not a foreign key", code=None):
        self.message = message
        self.code = code
        super().__init__(self.message)

    def __str__(self):
        if self.code:
            return f"[Error {self.code}]: {self.message}"
        return self.message



import argparse

from .abstraction import BaseCommand


class CommandManager:
    def __init__(self):
        self.parser = argparse.ArgumentParser(description="CLI for My DBEasyORM")
        self.subparsers = self.parser.add_subparsers(dest="command", help="Available commands")
        self.commands = {}

    def register_command(self, command_class: BaseCommand) -> None:
        """
        Register a command by its class.
        :param command_class: A class inheriting from BaseCommand.
        """
        command = command_class()
        subparser = self.subparsers.add_parser(command.name(), help=command.help())
        command.configure_arguments(subparser)
        self.commands[command.name()] = command

    def run(self) -> None:
        """Parse arguments and execute the appropriate command."""
        args = self.parser.parse_args()

        if args.command in self.commands:
            # Pass parsed arguments as a dictionary to the command's handler
            command_args = vars(args)
            self.commands[args.command].handle(**command_args)
        else:
            self.parser.print_help()



from setuptools import setup, find_packages

with open("README-PYPI.md", "r") as f:
    long_description = f.read()

setup(
    name="dbeasyorm",
    version="0.0.0",
    author="artur24814",
    author_email="artur24814@gmail.com",
    description="DBEasyORM is a lightweight and intuitive Object-Relational Mapping (ORM) library",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/artur24814/DBEasyORM",
    packages=find_packages(exclude=["docs", "tests", "app", ".github", "env"]),
    license="MIT",
    classifiers=[
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Database",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=["colorama >= 0.4.6"],
    extras_require={
        "dev": ["pytest>=8.3.4", "twine>=6.0.1"]
    },
    python_requires='>=3.8',
    project_urls={
        "Documentation": "https://dbeasyorm.readthedocs.io",
        "Source": "https://github.com/artur24814/DBEasyORM",
        "Bug Tracker": "https://github.com/artur24814/DBEasyORM/issues",
    },
)



from dbeasyorm.db.backends import DataBaseBackend
from dbeasyorm.models.model import Model


class MigrationDetecter:
    def __init__(self, db_backend: DataBaseBackend):
        self.db_backend = db_backend

    def get_detected_migrations(self, models: list) -> dict:
        db_schemas = self.get_database_schemas()
        return self.compare_schemas(models, db_schemas)

    def get_database_schemas(self) -> dict:
        self.db_backend.connect()
        return self.db_backend.get_database_schemas()

    def compare_schemas(self, models, db_schemas):
        migrations = {
            "create_tables": [],
            "drop_tables": [],
            "add_columns": [],
            "remove_columns": [],
        }

        for model in models:
            model_table_name = model.query_creator.get_table_name()
            if self.is_tables_exists_in_models_but_not_in_db(model_table_name, db_schemas.keys()):
                migrations["create_tables"].append(model)
                continue

            migrations = self.compare_cols_in_existing_tables(model_table_name, model, db_schemas, migrations)

        for table_name in db_schemas.keys():
            model_table_names = [model.query_creator.get_table_name() for model in models]
            if self.is_tables_exists_in_models_but_not_in_db(table_name, model_table_names):
                migrations["drop_tables"].append(table_name)

        return migrations

    def is_tables_exists_in_models_but_not_in_db(self, table_name: str, db_schemas_table_names: list) -> bool:
        return table_name not in db_schemas_table_names

    def compare_cols_in_existing_tables(self, table_name: str, model: Model, db_schemas: dict, migrations: dict) -> dict:
        db_columns = db_schemas[table_name]

        for field_name, field_instance in model._fields.items():
            if field_name not in db_columns:
                migrations["add_columns"].append((table_name, field_instance, model, db_columns))

        for column_name in db_columns.keys():
            if column_name not in list(model._fields.keys()):
                migrations["remove_columns"].append((table_name, column_name, model))

        return migrations

    def is_tables_exists_in_db_but_not_in_models(self, db_schemas_table_name: str, model_table_names: list) -> bool:
        return db_schemas_table_name not in model_table_names



from abc import ABC, abstractmethod
import argparse


class BaseCommand(ABC):
    """Abstract base class for all commands."""

    @abstractmethod
    def name(self) -> str:
        """Return the name of the command."""
        ...

    @abstractmethod
    def help(self) -> str:
        """Return the help description of the command."""
        ...

    @abstractmethod
    def configure_arguments(self, parser: argparse.ArgumentParser) -> None:
        """
        Configure arguments for the command.
        :param parser: Argument parser for the command.
        """
        ...

    @abstractmethod
    def handle(self, **kwargs) -> None:
        """
        Handle the command logic.
        :param kwargs: Parsed arguments passed as keyword arguments.
        """
        ...



from colorama import Fore
from dbeasyorm.config import _get_active_backend

from .migration_detecter import MigrationDetecter
from .migration_executor import MigrationExecutor
from .messages import print_success, print_line
from .model_classes_loader import ModelClassesLoader


class MigrationHandler:
    def __init__(self, config_file_path: str = None):
        self.db_backend = _get_active_backend(config_file_path) if config_file_path else _get_active_backend()
        self.migration_detec = MigrationDetecter(self.db_backend)
        self.migration_exec = MigrationExecutor(self.db_backend)
        self.models_loader = ModelClassesLoader()

    def update_database(self, loockup_folder: str, *args, **kwargs) -> None:
        print_line(Fore.GREEN, '=')
        models = self.models_loader.load_models(loockup_folder)
        detected_migration = self.migration_detec.get_detected_migrations(models)
        self.migration_exec.execute_detected_migration(detected_migration)
        print_success("Everything is up to date")
        print_line(Fore.GREEN, '=')



class Migration:
    @staticmethod
    def update_database(loockup_folder=None, id_migrations=None, restore=False):
        """
        Placeholder for the database update logic.
        :param loockup_folder: Path to the lookup folder
        :param id_migrations: String identifying specific migrations
        :param restore: Boolean flag to enable restore
        """
        print("Updating database with parameters:")
        print(f" - loockup_folder: {loockup_folder}")
        print(f" - id_migrations: {id_migrations}")
        print(f" - restore: {restore}")
